@@ -1,4 +1,4 @@
-// Copyright 2018 Chia Network Inc
+// Copyright 2020 Chia Network Inc
 
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,22 +15,12 @@
 #ifndef SRC_BLS_HPP_
 #define SRC_BLS_HPP_
 
-#include <vector>
-#include <map>
-#include <string>
-#include <stdexcept>
-
-#include "relic_conf.h"
-
-#if defined GMP && ARITH == GMP
-#include <gmp.h>
-#endif
-
 #include "privatekey.hpp"
 #include "util.hpp"
-
-#include "relic.h"
-#include "relic_test.h"
+#include "schemes.hpp"
+#include "elements.hpp"
+#include "hkdf.hpp"
+#include "hdkeys.hpp"
 
 namespace bls {
 
@@ -40,28 +30,14 @@ namespace bls {
  */
 class BLS {
  public:
-    // Order of g1, g2, and gt. Private keys are in {0, GROUP_ORDER}.
-    static const char GROUP_ORDER[];
     static const size_t MESSAGE_HASH_LEN = 32;
 
     // Initializes the BLS library (called automatically)
     static bool Init();
 
     static void SetSecureAllocator(Util::SecureAllocCallback allocCb, Util::SecureFreeCallback freeCb);
-    
-    /*
-    // Used for secure aggregation
-    static void HashPubKeys(
-            bn_t* output,
-            size_t numOutputs,
-            std::vector<uint8_t*> const &serPubKeys,
-            std::vector<size_t> const &sortedIndices);
-
-    static PublicKey DHKeyExchange(const PrivateKey& privKey, const PublicKey& pubKey);
-    */
 
     static void CheckRelicErrors();
-    static void CheckRelicErrorsInvalidArgument();
 };
 } // end namespace bls
 
